from __future__ import annotations

import sys
from collections import UserList
from enum import Enum
from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import Iterable
from typing import Mapping
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import Union

from qbittorrentapi._attrdict import AttrDict

if TYPE_CHECKING:
    from qbittorrentapi import Request

K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")

#: Type to define JSON.
JsonValueT = Union[
    None,
    int,
    str,
    bool,
    Sequence["JsonValueT"],
    Mapping[str, "JsonValueT"],
]
#: Type ``Any`` for ``kwargs`` parameters for API methods.
APIKwargsT = Any
#: Type for this API Client.
ClientT = TypeVar("ClientT", bound="Request")
#: Type for entry in List from API.
ListEntryT = TypeVar("ListEntryT", bound="ListEntry")
#: Type for List input to API method.
ListInputT = Iterable[Mapping[str, JsonValueT]]
#: Type for Files input to API method.
FilesToSendT = Mapping[str, Union[bytes, Tuple[str, bytes]]]


class APINames(str, Enum):
    """
    API namespaces for API endpoints.

    e.g ``torrents`` in ``http://localhost:8080/api/v2/torrents/addTrackers``
    """

    Authorization = "auth"
    Application = "app"
    Log = "log"
    Sync = "sync"
    Transfer = "transfer"
    Torrents = "torrents"
    RSS = "rss"
    Search = "search"
    EMPTY = ""


class TorrentState(str, Enum):
    """
    Torrent States as defined by qBittorrent.

    Definitions:
        - wiki: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-torrent-list>`_
        - code: `<https://github.com/qbittorrent/qBittorrent/blob/5dcc14153f046209f1067299494a82e5294d883a/src/base/bittorrent/torrent.h#L73>`_

    :Usage:
        >>> from qbittorrentapi import Client, TorrentState
        >>> client = Client()
        >>> # print torrent hashes for torrents that are downloading
        >>> for torrent in client.torrents_info():
        >>>     # check if torrent is downloading
        >>>     if torrent.state_enum.is_downloading:
        >>>         print(f'{torrent.hash} is downloading...')
        >>>     # the appropriate enum member can be directly derived
        >>>     state_enum = TorrentState(torrent.state)
        >>>     print(f'{torrent.hash}: {state_enum.value}')
    """

    ERROR = "error"
    MISSING_FILES = "missingFiles"
    UPLOADING = "uploading"
    PAUSED_UPLOAD = "pausedUP"
    QUEUED_UPLOAD = "queuedUP"
    STALLED_UPLOAD = "stalledUP"
    CHECKING_UPLOAD = "checkingUP"
    FORCED_UPLOAD = "forcedUP"
    ALLOCATING = "allocating"
    DOWNLOADING = "downloading"
    METADATA_DOWNLOAD = "metaDL"
    FORCED_METADATA_DOWNLOAD = "forcedMetaDL"
    PAUSED_DOWNLOAD = "pausedDL"
    QUEUED_DOWNLOAD = "queuedDL"
    FORCED_DOWNLOAD = "forcedDL"
    STALLED_DOWNLOAD = "stalledDL"
    CHECKING_DOWNLOAD = "checkingDL"
    CHECKING_RESUME_DATA = "checkingResumeData"
    MOVING = "moving"
    UNKNOWN = "unknown"

    @property
    def is_downloading(self) -> bool:
        """Returns ``True`` if the State is categorized as Downloading."""
        return self in {
            TorrentState.DOWNLOADING,
            TorrentState.METADATA_DOWNLOAD,
            TorrentState.FORCED_METADATA_DOWNLOAD,
            TorrentState.STALLED_DOWNLOAD,
            TorrentState.CHECKING_DOWNLOAD,
            TorrentState.PAUSED_DOWNLOAD,
            TorrentState.QUEUED_DOWNLOAD,
            TorrentState.FORCED_DOWNLOAD,
        }

    @property
    def is_uploading(self) -> bool:
        """Returns ``True`` if the State is categorized as Uploading."""
        return self in {
            TorrentState.UPLOADING,
            TorrentState.STALLED_UPLOAD,
            TorrentState.CHECKING_UPLOAD,
            TorrentState.QUEUED_UPLOAD,
            TorrentState.FORCED_UPLOAD,
        }

    @property
    def is_complete(self) -> bool:
        """Returns ``True`` if the State is categorized as Complete."""
        return self in {
            TorrentState.UPLOADING,
            TorrentState.STALLED_UPLOAD,
            TorrentState.CHECKING_UPLOAD,
            TorrentState.PAUSED_UPLOAD,
            TorrentState.QUEUED_UPLOAD,
            TorrentState.FORCED_UPLOAD,
        }

    @property
    def is_checking(self) -> bool:
        """Returns ``True`` if the State is categorized as Checking."""
        return self in {
            TorrentState.CHECKING_UPLOAD,
            TorrentState.CHECKING_DOWNLOAD,
            TorrentState.CHECKING_RESUME_DATA,
        }

    @property
    def is_errored(self) -> bool:
        """Returns ``True`` if the State is categorized as Errored."""
        return self in {TorrentState.MISSING_FILES, TorrentState.ERROR}

    @property
    def is_paused(self) -> bool:
        """Returns ``True`` if the State is categorized as Paused."""
        return self in {TorrentState.PAUSED_UPLOAD, TorrentState.PAUSED_DOWNLOAD}


TorrentStates = TorrentState


class TrackerStatus(int, Enum):
    """
    Tracker Statuses as defined by qBittorrent.

    Definitions:
        - wiki: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-torrent-trackers>`_
        - code: `<https://github.com/qbittorrent/qBittorrent/blob/5dcc14153f046209f1067299494a82e5294d883a/src/base/bittorrent/trackerentry.h#L42>`_

    :Usage:
        >>> from qbittorrentapi import Client, TrackerStatus
        >>> client = Client()
        >>> # print torrent hashes for torrents that are downloading
        >>> for torrent in client.torrents_info():
        >>>     for tracker in torrent.trackers:
        >>>         # display status for each tracker
        >>>         print(f"{torrent.hash[-6:]}: {TrackerStatus(tracker.status).display:>13} :{tracker.url}")
    """

    DISABLED = 0
    NOT_CONTACTED = 1
    WORKING = 2
    UPDATING = 3
    NOT_WORKING = 4

    @property
    def display(self) -> str:
        """Returns a descriptive display value for status."""
        return {
            TrackerStatus.DISABLED: "Disabled",
            TrackerStatus.NOT_CONTACTED: "Not contacted",
            TrackerStatus.WORKING: "Working",
            TrackerStatus.UPDATING: "Updating",
            TrackerStatus.NOT_WORKING: "Not working",
        }[self]


class ClientCache(Generic[ClientT]):
    """
    Caches the client.

    Subclass this for any object that needs access to the Client.
    """

    def __init__(self, *args: Any, client: ClientT, **kwargs: Any):
        self._client = client
        super().__init__(*args, **kwargs)


class Dictionary(AttrDict[V]):
    """Base definition of dictionary-like objects returned from qBittorrent."""

    def __init__(self, data: Mapping[str, JsonValueT] | None = None, **kwargs: Any):
        super().__init__(self._normalize(data or {}))
        # allows updating properties that aren't necessarily a part of the AttrDict
        self._setattr("_allow_invalid_attributes", True)

    @classmethod
    def _normalize(cls, data: Mapping[str, V] | T) -> AttrDict[V] | T:
        """Iterate through a dict converting any nested dicts to AttrDicts."""
        if isinstance(data, Mapping):
            return AttrDict({key: cls._normalize(value) for key, value in data.items()})
        return data


# Python 3.8 does not support UserList as a proper Generic
if sys.version_info < (3, 9):

    class List(UserList, Generic[ListEntryT]):
        """Base definition for list-like objects returned from qBittorrent."""

        def __init__(
            self,
            list_entries: ListInputT | None = None,
            entry_class: type[ListEntryT] | None = None,
            **kwargs: Any,
        ):
            super().__init__(
                [
                    entry_class(data=entry, **kwargs)
                    if entry_class is not None and isinstance(entry, Mapping)
                    else entry
                    for entry in list_entries or []
                ]
            )

else:

    class List(UserList[ListEntryT]):
        """Base definition for list-like objects returned from qBittorrent."""

        def __init__(
            self,
            list_entries: ListInputT | None = None,
            entry_class: type[ListEntryT] | None = None,
            **kwargs: Any,
        ):
            super().__init__(
                [
                    entry_class(data=entry, **kwargs)  # type: ignore[misc]
                    if entry_class is not None and isinstance(entry, Mapping)
                    else entry
                    for entry in list_entries or []
                ]
            )


class ListEntry(Dictionary[JsonValueT]):
    """Base definition for objects within a list returned from qBittorrent."""



from __future__ import annotations

from functools import wraps

from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.definitions import APIKwargsT
from qbittorrentapi.definitions import APINames
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.definitions import List
from qbittorrentapi.definitions import ListEntry
from qbittorrentapi.definitions import ListInputT


class LogPeer(ListEntry):
    """Item in :class:`LogPeersList`"""


class LogPeersList(List[LogPeer]):
    """Response for :meth:`~LogAPIMixIn.log_peers`"""

    def __init__(self, list_entries: ListInputT, client: LogAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=LogPeer)


class LogEntry(ListEntry):
    """Item in :class:`LogMainList`"""


class LogMainList(List[LogEntry]):
    """Response to :meth:`~LogAPIMixIn.log_main`"""

    def __init__(self, list_entries: ListInputT, client: LogAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=LogEntry)


class LogAPIMixIn(AppAPIMixIn):
    """
    Implementation of all ``Log`` API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> client.log_main(info=False)
        >>> client.log_peers()
    """

    @property
    def log(self) -> Log:
        """
        Allows for transparent interaction with Log endpoints.

        See Log class for usage.
        """
        if self._log is None:
            self._log = Log(client=self)
        return self._log

    def log_main(
        self,
        normal: bool | None = None,
        info: bool | None = None,
        warning: bool | None = None,
        critical: bool | None = None,
        last_known_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> LogMainList:
        """
        Retrieve the qBittorrent log entries. Iterate over returned object.

        :param normal: False to exclude ``normal`` entries
        :param info: False to exclude ``info`` entries
        :param warning: False to exclude ``warning`` entries
        :param critical: False to exclude ``critical`` entries
        :param last_known_id: only entries with an ID greater than this value will be returned
        """
        params = {
            "normal": None if normal is None else bool(normal),
            "info": None if info is None else bool(info),
            "warning": None if warning is None else bool(warning),
            "critical": None if critical is None else bool(critical),
            "last_known_id": last_known_id,
        }
        return self._get_cast(
            _name=APINames.Log,
            _method="main",
            params=params,
            response_class=LogMainList,
            **kwargs,
        )

    def log_peers(
        self,
        last_known_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> LogPeersList:
        """
        Retrieve qBittorrent peer log.

        :param last_known_id: only entries with an ID greater than this value will be
            returned
        """
        params = {"last_known_id": last_known_id}
        return self._get_cast(
            _name=APINames.Log,
            _method="peers",
            params=params,
            response_class=LogPeersList,
            **kwargs,
        )


class Log(ClientCache[LogAPIMixIn]):
    """
    Allows interaction with ``Log`` API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> # this is all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'log_' prepended)
        >>> log_list = client.log.main()
        >>> peers_list = client.log.peers(last_known_id="...")
        >>> # can also filter log down with additional attributes
        >>> log_info = client.log.main.info(last_known_id=1)
        >>> log_warning = client.log.main.warning(last_known_id=1)
    """

    def __init__(self, client: LogAPIMixIn):
        super().__init__(client=client)
        self.main = Log._Main(client=client)

    @wraps(LogAPIMixIn.log_peers)
    def peers(
        self,
        last_known_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> LogPeersList:
        return self._client.log_peers(last_known_id=last_known_id, **kwargs)

    class _Main(ClientCache["LogAPIMixIn"]):
        def _api_call(
            self,
            normal: bool | None = None,
            info: bool | None = None,
            warning: bool | None = None,
            critical: bool | None = None,
            last_known_id: str | int | None = None,
            **kwargs: APIKwargsT,
        ) -> LogMainList:
            return self._client.log_main(
                normal=normal,
                info=info,
                warning=warning,
                critical=critical,
                last_known_id=last_known_id,
                **kwargs,
            )

        def __call__(
            self,
            normal: bool | None = True,
            info: bool | None = True,
            warning: bool | None = True,
            critical: bool | None = True,
            last_known_id: str | int | None = None,
            **kwargs: APIKwargsT,
        ) -> LogMainList:
            return self._api_call(
                normal=normal,
                info=info,
                warning=warning,
                critical=critical,
                last_known_id=last_known_id,
                **kwargs,
            )

        def info(
            self,
            last_known_id: str | int | None = None,
            **kwargs: APIKwargsT,
        ) -> LogMainList:
            return self._api_call(last_known_id=last_known_id, **kwargs)

        def normal(
            self,
            last_known_id: str | int | None = None,
            **kwargs: APIKwargsT,
        ) -> LogMainList:
            return self._api_call(info=False, last_known_id=last_known_id, **kwargs)

        def warning(
            self,
            last_known_id: str | int | None = None,
            **kwargs: APIKwargsT,
        ) -> LogMainList:
            return self._api_call(
                info=False,
                normal=False,
                last_known_id=last_known_id,
                **kwargs,
            )

        def critical(
            self,
            last_known_id: str | int | None = None,
            **kwargs: APIKwargsT,
        ) -> LogMainList:
            return self._api_call(
                info=False,
                normal=False,
                warning=False,
                last_known_id=last_known_id,
                **kwargs,
            )



from __future__ import annotations

from functools import wraps
from json import dumps
from logging import Logger
from logging import getLogger
from typing import Any
from typing import Iterable
from typing import Mapping
from typing import Union

from qbittorrentapi.auth import AuthAPIMixIn
from qbittorrentapi.definitions import APIKwargsT
from qbittorrentapi.definitions import APINames
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.definitions import Dictionary
from qbittorrentapi.definitions import JsonValueT
from qbittorrentapi.definitions import List
from qbittorrentapi.definitions import ListEntry
from qbittorrentapi.definitions import ListInputT

logger: Logger = getLogger(__name__)


class ApplicationPreferencesDictionary(Dictionary[JsonValueT]):
    """
    Response for :meth:`~AppAPIMixIn.app_preferences`

    Definition: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-application-preferences>`_
    """  # noqa: E501


class BuildInfoDictionary(Dictionary[Union[str, int]]):
    """
    Response for :meth:`~AppAPIMixIn.app_build_info`

    Definition: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-build-info>`_
    """  # noqa: E501


class NetworkInterface(ListEntry):
    """Item in :class:`NetworkInterfaceList`"""


class NetworkInterfaceList(List[NetworkInterface]):
    """Response for :meth:`~AppAPIMixIn.app_network_interface_list`"""

    def __init__(self, list_entries: ListInputT, client: AppAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=NetworkInterface)


# only API response that's a list of strings...so just ignore the typing for now
class NetworkInterfaceAddressList(List[str]):  # type: ignore
    """Response for :meth:`~AppAPIMixIn.app_network_interface_address_list`"""

    def __init__(self, list_entries: Iterable[str], client: AppAPIMixIn | None = None):
        super().__init__(list_entries)  # type: ignore


class AppAPIMixIn(AuthAPIMixIn):
    """
    Implementation of all ``Application`` API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> client.app_version()
        >>> client.app_preferences()
    """

    @property
    def app(self) -> Application:
        """
        Allows for transparent interaction with Application endpoints.

        See Application class for usage.
        """
        if self._application is None:
            self._application = Application(client=self)
        return self._application

    application = app

    def app_version(self, **kwargs: APIKwargsT) -> str:
        """qBittorrent application version."""
        return self._get_cast(
            _name=APINames.Application,
            _method="version",
            response_class=str,
            **kwargs,
        )

    def app_web_api_version(self, **kwargs: APIKwargsT) -> str:
        """qBittorrent Web API version."""
        return self._get_cast(
            _name=APINames.Application,
            _method="webapiVersion",
            response_class=str,
            **kwargs,
        )

    app_webapiVersion = app_web_api_version

    def app_build_info(self, **kwargs: APIKwargsT) -> BuildInfoDictionary:
        """
        qBittorrent build info.

        This method was introduced with qBittorrent v4.2.0 (Web API v2.3).
        """
        return self._get_cast(
            _name=APINames.Application,
            _method="buildInfo",
            response_class=BuildInfoDictionary,
            version_introduced="2.3",
            **kwargs,
        )

    app_buildInfo = app_build_info

    def app_shutdown(self, **kwargs: APIKwargsT) -> None:
        """Shutdown qBittorrent."""
        self._post(_name=APINames.Application, _method="shutdown", **kwargs)

    def app_preferences(self, **kwargs: APIKwargsT) -> ApplicationPreferencesDictionary:
        """Retrieve qBittorrent application preferences."""
        return self._get_cast(
            _name=APINames.Application,
            _method="preferences",
            response_class=ApplicationPreferencesDictionary,
            **kwargs,
        )

    def app_set_preferences(
        self,
        prefs: Mapping[str, Any] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Set one or more preferences in qBittorrent application.

        :param prefs: dictionary of preferences to set
        """
        data = {"json": dumps(prefs, separators=(",", ":"))}
        self._post(
            _name=APINames.Application,
            _method="setPreferences",
            data=data,
            **kwargs,
        )

    app_setPreferences = app_set_preferences

    def app_default_save_path(self, **kwargs: APIKwargsT) -> str:
        """The default path where torrents are saved."""
        return self._get_cast(
            _name=APINames.Application,
            _method="defaultSavePath",
            response_class=str,
            **kwargs,
        )

    app_defaultSavePath = app_default_save_path

    def app_network_interface_list(self, **kwargs: APIKwargsT) -> NetworkInterfaceList:
        """
        The list of network interfaces on the host.

        This method was introduced with qBittorrent v4.2.0 (Web API v2.3).
        """
        return self._get_cast(
            _name=APINames.Application,
            _method="networkInterfaceList",
            response_class=NetworkInterfaceList,
            version_introduced="2.3",
            **kwargs,
        )

    app_networkInterfaceList = app_network_interface_list

    def app_network_interface_address_list(
        self,
        interface_name: str = "",
        **kwargs: APIKwargsT,
    ) -> NetworkInterfaceAddressList:
        """
        The addresses for a network interface; omit name for all addresses.

        This method was introduced with qBittorrent v4.2.0 (Web API v2.3).

        :param interface_name: Name of interface to retrieve addresses for
        """
        data = {"iface": interface_name}
        return self._post_cast(
            _name=APINames.Application,
            _method="networkInterfaceAddressList",
            data=data,
            response_class=NetworkInterfaceAddressList,
            version_introduced="2.3",
            **kwargs,
        )

    app_networkInterfaceAddressList = app_network_interface_address_list


class Application(ClientCache[AppAPIMixIn]):
    """
    Allows interaction with ``Application`` API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> # these are all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'app_' prepended)
        >>> webapiVersion = client.application.webapiVersion
        >>> web_api_version = client.application.web_api_version
        >>> app_web_api_version = client.application.web_api_version
        >>> # access and set preferences as attributes
        >>> is_dht_enabled = client.application.preferences.dht
        >>> # supports sending a just subset of preferences to update
        >>> client.application.preferences = dict(dht=(not is_dht_enabled))
        >>> prefs = client.application.preferences
        >>> prefs["web_ui_clickjacking_protection_enabled"] = True
        >>> client.app.preferences = prefs
        >>>
        >>> client.application.shutdown()
    """

    @property
    @wraps(AppAPIMixIn.app_version)
    def version(self) -> str:
        return self._client.app_version()

    @property
    @wraps(AppAPIMixIn.app_web_api_version)
    def web_api_version(self) -> str:
        return self._client.app_web_api_version()

    webapiVersion = web_api_version

    @property
    @wraps(AppAPIMixIn.app_build_info)
    def build_info(self) -> BuildInfoDictionary:
        return self._client.app_build_info()

    buildInfo = build_info

    @wraps(AppAPIMixIn.app_shutdown)
    def shutdown(self, **kwargs: APIKwargsT) -> None:
        self._client.app_shutdown(**kwargs)

    @property
    @wraps(AppAPIMixIn.app_preferences)
    def preferences(self) -> ApplicationPreferencesDictionary:
        return self._client.app_preferences()

    @preferences.setter
    @wraps(AppAPIMixIn.app_set_preferences)
    def preferences(self, value: Mapping[str, Any]) -> None:
        self.set_preferences(prefs=value)

    @wraps(AppAPIMixIn.app_set_preferences)
    def set_preferences(
        self,
        prefs: Mapping[str, Any] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        self._client.app_set_preferences(prefs=prefs, **kwargs)

    setPreferences = set_preferences

    @property
    @wraps(AppAPIMixIn.app_default_save_path)
    def default_save_path(self) -> str:
        return self._client.app_default_save_path()

    defaultSavePath = default_save_path

    @property
    @wraps(AppAPIMixIn.app_network_interface_list)
    def network_interface_list(self) -> NetworkInterfaceList:
        return self._client.app_network_interface_list()

    networkInterfaceList = network_interface_list

    @wraps(AppAPIMixIn.app_network_interface_address_list)
    def network_interface_address_list(
        self,
        interface_name: str = "",
        **kwargs: APIKwargsT,
    ) -> NetworkInterfaceAddressList:
        return self._client.app_network_interface_address_list(
            interface_name=interface_name,
            **kwargs,
        )

    networkInterfaceAddressList = network_interface_address_list



import sys

import pytest

from qbittorrentapi import APINames
from qbittorrentapi import NotFound404Error
from qbittorrentapi.search import SearchCategoriesList
from qbittorrentapi.search import SearchJobDictionary
from qbittorrentapi.search import SearchPluginsList
from qbittorrentapi.search import SearchResultsDictionary
from qbittorrentapi.search import SearchStatusesList
from tests.utils import check
from tests.utils import retry

PLUGIN_NAME = "yts"
PLUGIN_URL = "https://raw.githubusercontent.com/khensolomon/leyts/master/yts.py"


@pytest.mark.skipif(sys.version_info < (3, 9), reason="removeprefix not in 3.8")
def test_methods(client):
    namespace = APINames.Search
    all_dotted_methods = set(dir(getattr(client, namespace)))

    for meth in [meth for meth in dir(client) if meth.startswith(f"{namespace}_")]:
        assert meth.removeprefix(f"{namespace}_") in all_dotted_methods


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "update_func", ["search_update_plugins", "search.update_plugins"]
)
def test_update_plugins(client, update_func):
    client.func(update_func)()
    check(
        lambda: any(
            entry.message.startswith("Updating plugin ")
            or entry.message == "All plugins are already up to date."
            or entry.message.endswith("content was not found at the server (404)")
            for entry in reversed(client.log.main())
        ),
        True,
    )


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "update_func", ["search_update_plugins", "search.update_plugins"]
)
def test_update_plugins_not_implemented(client, update_func):
    with pytest.raises(NotImplementedError):
        client.func(update_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "search_func, enable_func",
    (
        ["search_plugins", "search_enable_plugin"],
        ["search.plugins", "search.enable_plugin"],
    ),
)
def test_enable_plugin(client, search_func, enable_func):
    def get_plugins():
        try:
            return client.func(search_func)()
        except TypeError:
            return client.func(search_func)

    @retry()
    def enable_plugin():
        assert isinstance(get_plugins(), SearchPluginsList)
        client.func(enable_func)(
            plugins=(p["name"] for p in get_plugins()), enable=False
        )
        check(
            lambda: (p["enabled"] for p in get_plugins()),
            True,
            reverse=True,
            negate=True,
        )
        client.func(enable_func)(
            plugins=(p["name"] for p in get_plugins()), enable=True
        )
        check(
            lambda: (p["enabled"] for p in get_plugins()),
            False,
            reverse=True,
            negate=True,
        )

    enable_plugin()


@pytest.mark.skipif_before_api_version("2.1.1")
def test_plugins_slice(client):
    assert isinstance(client.search_plugins()[1:2], SearchPluginsList)


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "enable_func", ["search_enable_plugin", "search.enable_plugin"]
)
def test_enable_plugin_not_implemented(client, enable_func):
    with pytest.raises(NotImplementedError):
        client.func(enable_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "install_func, uninstall_func",
    (
        ["search_install_plugin", "search_uninstall_plugin"],
        ["search.install_plugin", "search.uninstall_plugin"],
    ),
)
def test_install_uninstall_plugin(client, install_func, uninstall_func):
    @retry()
    def install_plugin():
        client.func(install_func)(sources=PLUGIN_URL)
        check(
            lambda: (p.name for p in client.search.plugins),
            PLUGIN_NAME,
            reverse=True,
        )

    @retry()
    def uninstall_plugin():
        client.func(uninstall_func)(names=PLUGIN_NAME)
        check(
            lambda: (p.name for p in client.search.plugins),
            PLUGIN_NAME,
            reverse=True,
            negate=True,
        )

    install_plugin()
    uninstall_plugin()


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "install_func, uninstall_func",
    (
        ["search_install_plugin", "search_uninstall_plugin"],
        ["search.install_plugin", "search.uninstall_plugin"],
    ),
)
def test_install_uninstall_plugin_not_implemented(client, install_func, uninstall_func):
    with pytest.raises(NotImplementedError):
        client.func(install_func)()
    with pytest.raises(NotImplementedError):
        client.func(uninstall_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.skipif_after_api_version("2.6")
@pytest.mark.parametrize("categories_func", ["search_categories", "search.categories"])
def test_categories(client, categories_func):
    assert isinstance(client.func(categories_func)(), SearchCategoriesList)
    assert isinstance(client.func(categories_func)()[1:2], SearchCategoriesList)
    check(lambda: client.func(categories_func)(), "All categories", reverse=True)


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize("categories_func", ["search_categories", "search.categories"])
def test_categories_not_implemented(client, categories_func):
    with pytest.raises(NotImplementedError):
        client.func(categories_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "start_func, status_func, results_func, stop_func, delete_stop",
    [
        (
            "search_start",
            "search_status",
            "search_results",
            "search_stop",
            "search_delete",
        ),
        (
            "search.start",
            "search.status",
            "search.results",
            "search.stop",
            "search.delete",
        ),
    ],
)
def test_search(client, start_func, status_func, results_func, stop_func, delete_stop):
    job = client.func(start_func)(pattern="Ubuntu", plugins="enabled", category="all")

    statuses = client.func(status_func)(search_id=job["id"])
    assert statuses[0]["status"] == "Running"
    assert isinstance(job, SearchJobDictionary)
    assert isinstance(statuses, SearchStatusesList)

    results = client.func(results_func)(search_id=job["id"], limit=1)
    assert isinstance(results, SearchResultsDictionary)
    results = job.results()
    assert isinstance(results, SearchResultsDictionary)

    client.func(stop_func)(search_id=job["id"])
    check(lambda: client.func(status_func)(search_id=job["id"])[0]["status"], "Stopped")

    client.func(delete_stop)(search_id=job["id"])
    statuses = client.func(status_func)()
    assert not statuses


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize("status_func", ["search_status", "search.status"])
def test_statuses_slice(client, status_func):
    assert isinstance(client.func(status_func)()[1:2], SearchStatusesList)


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func",
    [
        "search_start",
        "search_status",
        "search_results",
        "search_stop",
        "search_delete",
        "search.start",
        "search.status",
        "search.results",
        "search.stop",
        "search.delete",
    ],
)
def test_search_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        client.func(client_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
@pytest.mark.parametrize(
    "stop_func, start_func",
    [("search_stop", "search_start"), ("search.stop", "search.start")],
)
def test_stop(client, stop_func, start_func):
    job = client.func(start_func)(pattern="Ubuntu", plugins="enabled", category="all")
    check(lambda: client.search.status(search_id=job["id"])[0]["status"], "Running")

    client.func(stop_func)(search_id=job.id)
    check(lambda: client.search.status(search_id=job["id"])[0]["status"], "Stopped")

    job = client.func(start_func)(pattern="Ubuntu", plugins="enabled", category="all")
    check(lambda: client.search.status(search_id=job["id"])[0]["status"], "Running")
    job.stop()
    check(lambda: client.search.status(search_id=job["id"])[0]["status"], "Stopped")


@pytest.mark.skipif_after_api_version("2.1.1")
@pytest.mark.parametrize(
    "client_func", ["search_stop", "search_start", "search.stop", "search.start"]
)
def test_stop_not_implemented(client, client_func):
    with pytest.raises(NotImplementedError):
        client.func(client_func)()


@pytest.mark.skipif_before_api_version("2.1.1")
def test_delete(client):
    job = client.search_start(pattern="Ubuntu", plugins="enabled", category="all")
    job.delete()
    with pytest.raises(NotFound404Error):
        job.status()


@pytest.mark.skipif_after_api_version("2.1.1")
def test_delete_not_implemented(client):
    with pytest.raises(NotImplementedError):
        client.search_stop(search_id=100)
